# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2021 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""reana-client output print configuration."""

import click

from reana_client.config import (
    PRINTER_COLOUR_ERROR,
    PRINTER_COLOUR_INFO,
    PRINTER_COLOUR_SUCCESS,
    PRINTER_COLOUR_WARNING,
)


def display_message(msg, msg_type=None, indented=False):
    """Display messages in console.

    :param msg: Message to display
    :param msg_type: Type of message (info/note/warning/error)
    :param indented: Message indented or not
    :type msg: str
    :type msg_type: str
    :type indented: bool
    """
    msg_color_map = {
        "success": PRINTER_COLOUR_SUCCESS,
        "warning": PRINTER_COLOUR_WARNING,
        "error": PRINTER_COLOUR_ERROR,
        "info": PRINTER_COLOUR_INFO,
    }
    msg_color = msg_color_map.get(msg_type, "")

    if msg_type == "info":
        if indented:
            click.secho(
                "  -> {}: ".format(msg_type.upper()), bold=True, nl=False, fg=msg_color,
            )
            click.secho("{}".format(msg), nl=True)
        else:
            click.secho("==> ", bold=True, nl=False)
            click.secho("{}".format(msg), bold=True, nl=True)
    elif msg_type in ["error", "warning", "success"]:
        prefix_tpl = "  -> {}: " if indented else "==> {}: "
        click.secho(
            prefix_tpl.format(msg_type.upper()),
            bold=True,
            nl=False,
            err=msg_type == "error",
            fg=msg_color,
        )
        click.secho("{}".format(msg), bold=False, err=msg_type == "error", nl=True)
    else:
        click.secho("{}".format(msg), nl=True)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2021 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client validate environments tests."""

import pytest

from reana_client.errors import EnvironmentValidationError
from reana_client.validation.environments import SerialEnvironmentValidator


@pytest.mark.parametrize(
    "image, output, exit_",
    [
        ("foo:bar", "has the correct format", False),
        ("foo/bar:baz", "has the correct format", False),
        ("foo/bar", "not have an explicit tag", False),
        ("foo/bar:", "tag is not recommended", False),
        ("foo/bar:latest", "tag is not recommended", False),
        ("foo:master", "tag is not recommended", False),
        ("foo:bar:baz", "has invalid tag", True),
        ("foo:bar:", "has invalid tag", True),
    ],
)
def test_validate_environment_image_tag(image, output, exit_):
    """Validate workflow environment image tags."""
    validator = SerialEnvironmentValidator()
    if exit_:
        with pytest.raises(EnvironmentValidationError) as e:
            validator._validate_image_tag(image)
        assert output in str(e)
    else:
        validator._validate_image_tag(image)
        assert output in validator.messages.pop()["message"]


@pytest.mark.parametrize(
    "image, tag, full_image_name",
    [
        ("foo", "bar", "foo:bar"),
        ("foo", "", "foo"),
        ("foo", None, "foo"),
        ("foo", "latest", "foo:latest"),
    ],
)
def test_get_full_image_name(image, tag, full_image_name):
    validator = SerialEnvironmentValidator()
    assert validator._get_full_image_name(image, tag) == full_image_name



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2021 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration for REANA client."""

from __future__ import absolute_import, print_function

import pytest


@pytest.fixture()
def create_yaml_workflow_schema():
    """Return dummy yaml workflow schema."""
    reana_yaml_schema = """
        version: 0.7.2
        inputs:
          files:
            - code/helloworld.py
            - inputs/names.txt
          parameters:
            sleeptime: 2
            inputfile: inputs/names.txt
            helloworld: code/helloworld.py
            outputfile: outputs/greetings.txt
        outputs:
          files:
           - outputs/greetings.txt
        workflow:
          type: serial
          specification:
            steps:
              - environment: 'python:2.7'
                commands:
                  - python "${helloworld}" --sleeptime ${sleeptime} \
                  --inputfile "${inputfile}" --outputfile "${outputfile}"
        """
    return reana_yaml_schema


@pytest.fixture()
def create_cwl_yaml_workflow_schema():
    """Return dummy cwl workflow schema."""
    reana_cwl_yaml_schema = """
        version: 0.7.2
        workflow:
          type: cwl
          file: main.cwl
        outputs:
          files:
            - foo/bar
        """
    return reana_cwl_yaml_schema


@pytest.fixture()
def cwl_workflow_spec_step():
    """Return dummy CWL workflow loaded spec step."""
    cwl_workflow_spec_step = """
        #!/usr/bin/env cwl-runner

        cwlVersion: v1.0
        class: Workflow

        inputs:
          outputfile:
            type: string
            inputBinding:
              prefix: --outputfile

        outputs:
          result:
            type: File
            outputSource: first/result

        steps:
          first:
            run: test.tool
            in:
              outputfile: outputfile
            out: [result]
        """
    return cwl_workflow_spec_step


@pytest.fixture()
def cwl_workflow_spec_correct_input_param():
    """Return correct dummy CWL workflow loaded spec."""
    cwl_workflow_spec = """
        #!/usr/bin/env cwl-runner

        cwlVersion: v1.0
        class: CommandLineTool

        baseCommand: python

        inputs:
          outputfile:
            type: string
            inputBinding:
              prefix: --outputfile

        outputs:
          result:
            type: File
            outputBinding:
              glob: $(inputs.outputfile)
        """
    return cwl_workflow_spec


@pytest.fixture()
def cwl_workflow_spec_wrong_input_param():
    """Return wrong dummy CWL workflow loaded spec."""
    cwl_workflow_spec = """
        #!/usr/bin/env cwl-runner

        cwlVersion: v1.0
        class: CommandLineTool

        baseCommand: python

        inputs:
          xoutputfile:  # wrong input param
            type: string
            inputBinding:
              prefix: --outputfile

        outputs:
          result:
            type: File
            outputBinding:
              glob: $(inputs.outputfile)
        """
    return cwl_workflow_spec



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA-Client."""

from __future__ import absolute_import, print_function

from .version import __version__

__all__ = ("__version__",)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA-Client errors."""


class FileUploadError(Exception):
    """File upload didn't succeed."""


class FileDeletionError(Exception):
    """File deletion didn't succeed."""


class EnvironmentValidationError(Exception):
    """REANA workflow environment validation didn't succeed."""


class ParameterValidationError(Exception):
    """REANA workflow parameter validation didn't succeed."""



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2019 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client API utils."""


def get_path_from_operation_id(paths_dict, operation_id):
    """Find API path based on operation id."""
    paths = paths_dict.keys()
    for path in paths:
        methods = paths_dict[path].keys()
        for method in methods:
            if paths_dict[path][method]["operationId"] == operation_id:
                return path
    return None



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2019 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client secrets tests."""

import pytest
from bravado.exception import HTTPError
from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client

from reana_client.cli import cli


def test_secrets_list_server_not_reachable():
    """Test list secrets when not connected to any cluster."""
    message = "REANA client is not connected to any REANA cluster."
    reana_token = "000000"
    runner = CliRunner()
    result = runner.invoke(cli, ["secrets-list", "-t", reana_token])
    assert result.exit_code == 1
    assert message in result.output


def test_secrets_list_server_no_token():
    """Test list secrets when access token is not set."""
    message = "Please provide your access token"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["secrets-list"])
    assert result.exit_code == 1
    assert message in result.output


def test_secrets_list_ok():
    """Test list secrets successfull."""
    status_code = 200
    response = [{"name": "password", "type": "env"}]
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["secrets-list", "-t", reana_token])
            assert result.exit_code == 0
            assert "password" in result.output
            assert "env" in result.output


@pytest.mark.parametrize("secret", ["USER=reanauser", "USER=reana=user"])
def test_secrets_add(secret):
    """Test secrets add."""
    status_code = 201
    reana_token = "000000"
    secret_file = "file.txt"
    response = [secret_file]
    env = {"REANA_SERVER_URL": "localhost"}
    message = "were successfully uploaded."
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_http_response.raw_bytes = str(response).encode()
    mock_response = response
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            with runner.isolated_filesystem():
                with open(secret_file, "w") as f:
                    f.write("test")
                result = runner.invoke(
                    cli,
                    [
                        "secrets-add",
                        "-t",
                        reana_token,
                        "--file",
                        secret_file,
                        "--env",
                        secret,
                    ],
                )
                assert result.exit_code == 0
                assert message in result.output


@pytest.mark.parametrize("secret", ["wrongformat", "PASS:123"])
def test_secrets_add_wrong_format(secret):
    """Test adding secrets with wrong format."""
    reana_token = "000000"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    message = 'For literal strings use "SECRET_NAME=VALUE" format'

    result = runner.invoke(cli, ["secrets-add", "-t", reana_token, "--env", secret])
    assert result.exit_code == 1
    assert message in result.output


def test_secrets_add_already_exist():
    """Test adding secrets when they already exist."""
    status_code = 409
    reana_token = "000000"
    env = {"REANA_SERVER_URL": "localhost"}
    message = "One of the secrets already exists. No secrets were added."
    mock_http_response = Mock(
        status_code=status_code,
        reason="Conflict",
        json=Mock(return_value={"message": "Conflict"}),
    )
    rs_api_client_mock = Mock()
    rs_api_client_mock.api.add_secrets = Mock(side_effect=HTTPError(mock_http_response))
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.current_rs_api_client", rs_api_client_mock):
            result = runner.invoke(
                cli, ["secrets-add", "-t", reana_token, "--env", "USER=reanauser"]
            )
            assert message in result.output
            assert result.exit_code == 1


def test_secrets_delete():
    """Test secrets delete."""
    status_code = 200
    reana_token = "000000"
    secret_file = "file.txt"
    response = ["file.txt"]
    message = "were successfully deleted."
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["secrets-delete", "-t", reana_token, secret_file]
            )
            assert result.exit_code == 0
            assert message in result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2020 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client CLI API docs generation."""

import click

from reana_client.cli import cli

CLI_API_URL = "https://reana-client.readthedocs.io/en/latest/#cli-api"


def _print_code_block(content, lang=""):
    print("```{}".format(lang))
    print(content)
    print("```")


def generate_cli_docs():
    """Generate Markdown friendly CLI API documentation."""
    print("# reana-client CLI API\n")
    print("The complete `reana-client` CLI API reference guide is available here:\n")
    print(f"- [{CLI_API_URL}]({CLI_API_URL})\n")
    with click.Context(cli) as ctx:
        _print_code_block(cli.get_help(ctx), lang="console")

    for cmd_group in cli.cmd_groups:
        print("\n## {}".format(cmd_group.help))
        for cmd_obj in cmd_group.commands.values():
            print("\n### {}\n".format(cmd_obj.name))
            print(cmd_obj.help)


if __name__ == "__main__":
    generate_cli_docs()



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client workflow tests."""

import json

import pytest
import yaml
from click.testing import CliRunner
from mock import Mock, patch
from pytest_reana.test_utils import make_mock_api_client
from reana_commons.config import INTERACTIVE_SESSION_TYPES

from reana_client.api.client import create_workflow_from_json
from reana_client.cli import cli
from reana_client.utils import get_workflow_status_change_msg


def test_workflows_server_not_connected():
    """Test workflows command when server is not connected."""
    runner = CliRunner()
    reana_token = "000000"
    result = runner.invoke(cli, ["list", "-t", reana_token])
    message = "REANA client is not connected to any REANA cluster."
    assert message in result.output
    assert result.exit_code == 1


def test_workflows_no_token():
    """Test workflows command when token is not set."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["list"])
    message = "Please provide your access token by using the -t"
    assert result.exit_code == 1
    assert message in result.output


def test_workflows_server_ok():
    """Test workflows command when server is reachable."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {
                    "run_started_at": "2018-06-13T09:47:40.28223",
                    "run_finished_at": "2018-06-13T10:30:03.70303",
                },
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost", "REANA_WORKON": "mytest.1"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "-t", reana_token, "--include-progress"]
            )
            message = "RUN_NUMBER"
            assert result.exit_code == 0
            assert message in result.output


def test_workflows_sorting():
    """Test workflows sorting."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "progress": {
                    "run_started_at": "2018-06-13T09:47:40.28223",
                    "run_finished_at": "2018-06-13T10:30:03.70303",
                },
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            },
            {
                "status": "running",
                "created": "2018-06-13T09:55:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.2",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a2",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            },
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost", "REANA_WORKON": "mytest.1"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "-t", reana_token, "--sort", "run_number"]
            )
            message = (
                "mytest   2            2018-06-13T09:55:35.66097   -"
                "                           -"
                "                           running\n"
                "mytest   1            2018-06-13T09:47:35.66097   "
                "2018-06-13T09:47:40.28223   2018-06-13T10:30:03.70303"
                "   running"
            )
            assert result.exit_code == 0
            assert message in result.output


def test_workflows_sessions():
    """Test list command for getting interactive sessions."""
    response = {
        "items": [
            {
                "created": "2019-03-19T14:37:58",
                "id": "29136cd0-b259-4d48-8c1e-afe3572df408",
                "name": "workflow.1",
                "session_type": "jupyter",
                "session_uri": "/29136cd0-b259-4d48-8c1e-afe3572df408",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "status": "created",
                "user": "00000000-0000-0000-0000-000000000000",
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost", "REANA_WORKON": "mytest.1"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-t", reana_token, "--sessions"])
            message = "RUN_NUMBER"
            assert result.exit_code == 0
            assert message in result.output


def test_workflows_valid_json():
    """Test workflows command with --json and -v flags."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-v", "-t", reana_token, "--json"])
            assert result.exit_code == 0


def test_workflows_include_progress():
    """Test workflows command with --include-progress flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {
                    "total": {"job_ids": [], "total": 5},
                    "running": {"job_ids": [], "total": 3},
                    "finished": {"job_ids": [], "total": 2},
                    "failed": {"job_ids": [], "total": 0},
                },
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "--include-progress", "-t", reana_token]
            )
            assert result.exit_code == 0
            assert "PROGRESS" in result.output
            assert "2/5" in result.output


def test_workflows_without_include_progress():
    """Test workflows command without --include-progress flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
                "progress": {
                    "run_started_at": "2021-05-10T12:55:04",
                    "run_finished_at": "2021-05-10T12:55:23",
                },
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-t", reana_token])
            assert result.exit_code == 0
            assert "PROGRESS" not in result.output
            assert "STARTED" in result.output
            assert "2021-05-10T12:55:04" in result.output


def test_workflows_include_workspace_size():
    """Test workflows command with --include-workspace-size flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"human_readable": "15.97 MiB", "raw": 16741346},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "--include-workspace-size", "-t", reana_token]
            )
            assert result.exit_code == 0
            assert "SIZE" in result.output
            assert "16741346" in result.output


def test_workflows_without_include_workspace_size():
    """Test workflows command without --include-workspace-size flag."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"human_readable": "", "raw": -1},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["list", "-t", reana_token])
            assert result.exit_code == 0
            assert "SIZE" not in result.output


def test_workflows_format():
    """Test workflows command with --format."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            },
            {
                "status": "failed",
                "created": "2018-06-14T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.2",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a2",
                "size": {"raw": 0, "human_readable": "0 Bytes"},
            },
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    _format = "status=failed"
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                ["list", "-t", reana_token, "--json", '--format="{}"'.format(_format)],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert "status" in json_response[0]
            assert json_response[0]["status"] == "failed"


def test_workflows_filter():
    """Test workflows command with --filter."""
    response = {
        "items": [
            {
                "status": "running",
                "created": "2018-06-13T09:47:35.66097",
                "user": "00000000-0000-0000-0000-000000000000",
                "name": "mytest.1",
                "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
                "size": {"raw": 340, "human_readable": "340 Bytes"},
            }
        ]
    }
    status_code = 200
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    runner = CliRunner(env=env)
    filter = "status=running"
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["list", "-t", reana_token, "--json", "--filter", filter],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert "status" in json_response[0]
            assert "running" in json_response[0]["status"]


def test_workflow_create_failed():
    """Test workflow create when creation fails."""
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["create"])
    message = "ERROR: No REANA specification file (reana.yaml) found"
    assert message in result.output
    assert result.exit_code == 1


def test_workflow_create_successful(create_yaml_workflow_schema):
    """Test workflow create when creation is successfull."""
    status_code = 201
    response = {
        "message": "The workflow has been successfully created.",
        "workflow_id": "cdcf48b1-c2f3-4693-8230-b066e088c6ac",
        "workflow_name": "mytest.1",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            with runner.isolated_filesystem():
                with open("reana.yaml", "w") as f:
                    f.write(create_yaml_workflow_schema)
                result = runner.invoke(
                    cli, ["create", "-t", reana_token, "--skip-validation"]
                )
                assert result.exit_code == 0
                assert response["workflow_name"] in result.output


def test_workflow_create_not_valid_name(create_yaml_workflow_schema):
    """Test workflow create when creation is successfull."""
    env = {"REANA_SERVER_URL": "localhost"}
    illegal_workflow_name = "workflow.name"
    runner = CliRunner(env=env)
    result = runner.invoke(cli, ["create", "-n", illegal_workflow_name])
    message = 'Workflow name {} contains illegal character "{}"'.format(
        illegal_workflow_name, "."
    )
    assert message in result.output
    assert result.exit_code == 1


def test_create_workflow_from_json(create_yaml_workflow_schema):
    """Test create workflow from json specification."""
    status_code = 201
    response = {
        "message": "The workflow has been successfully created.",
        "workflow_id": "cdcf48b1-c2f3-4693-8230-b066e088c6ac",
        "workflow_name": "mytest.1",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    reana_token = "000000"
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    workflow_json = yaml.load(create_yaml_workflow_schema, Loader=yaml.FullLoader)
    with patch.dict("os.environ", env):
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = create_workflow_from_json(
                workflow_json=workflow_json["workflow"],
                name=response["workflow_name"],
                access_token=reana_token,
                parameters=workflow_json["inputs"],
                workflow_engine="serial",
            )
            assert response["workflow_name"] == result["workflow_name"]
            assert response["message"] == result["message"]


@pytest.mark.parametrize(
    "status",
    ["created", "running", "finished", "failed", "deleted", "stopped", "queued"],
)
def test_workflow_start_successful(status):
    """Test workflow start when creation is successfull."""
    workflow_name = "mytest.1"
    response = {
        "status": status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000",
    }
    status_code = 200
    reana_token = "000000"
    expected_message = get_workflow_status_change_msg(workflow_name, status)
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli, ["start", "-t", reana_token, "-w", response["workflow_name"]]
            )
            assert result.exit_code == 0
            assert expected_message in result.output


@pytest.mark.parametrize(
    "initial_status, final_status, exit_code",
    [("running", "finished", 0), ("running", "failed", 1), ("running", "stopped", 1)],
)
def test_workflow_start_follow(initial_status, final_status, exit_code):
    """Test start workflow with follow flag."""
    workflow_name = "mytest.1"
    initial_reponse = {
        "status": initial_status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000",
        "size": {"raw": 0, "human_readable": "0 Bytes"},
    }
    final_reponse = {
        "status": final_status,
        "message": "Server message",
        "id": "256b25f4-4cfb-4684-b7a8-73872ef455a1",
        "workflow_name": workflow_name,
        "user": "00000000-0000-0000-0000-000000000000",
        "size": {"raw": 340, "human_readable": "340 Bytes"},
    }
    initial_expected_message = get_workflow_status_change_msg(
        workflow_name, initial_status
    )
    final_Expected_message = get_workflow_status_change_msg(workflow_name, final_status)
    status_code = 200
    mock_http_response = Mock()
    mock_http_response.status_code = status_code
    mock_api_client = Mock()
    mock_start_workflow_result = Mock(
        return_value=(mock_http_response, mock_http_response)
    )
    mock_api_client.api.start_workflow.return_value = Mock(
        result=mock_start_workflow_result
    )
    mock_get_workflow_status_result = Mock(
        side_effect=[
            (initial_reponse, mock_http_response),
            (final_reponse, mock_http_response),
        ]
    )
    mock_api_client.api.get_workflow_status.return_value = Mock(
        result=mock_get_workflow_status_result
    )
    reana_token = "000000"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch("reana_client.api.client.current_rs_api_client", mock_api_client):
            result = runner.invoke(
                cli, ["start", "-t", reana_token, "-w", workflow_name, "--follow"]
            )
            assert result.exit_code == exit_code
            assert initial_expected_message in result.output
            assert final_Expected_message in result.output


def test_workflows_validate(create_yaml_workflow_schema):
    """Test validation of REANA specifications file."""
    message = "Valid REANA specification file"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    with runner.isolated_filesystem():
        with open("reana.yaml", "w") as f:
            f.write(create_yaml_workflow_schema)
        result = runner.invoke(cli, ["validate", "--file", "reana.yaml"],)
        assert result.exit_code == 0
        assert message in result.output


def test_get_workflow_status_ok():
    """Test workflow status."""
    status_code = 200
    response = {
        "created": "2018-10-29T12:50:12",
        "id": "4e576cf9-a946-4346-9cde-7712f8dcbb3f",
        "logs": "",
        "name": "workflow.5",
        "progress": {
            "current_command": None,
            "current_step_name": None,
            "failed": {"job_ids": [], "total": 0},
            "finished": {"job_ids": [], "total": 0},
            "run_started_at": "2018-10-29T12:51:04",
            "running": {"job_ids": [], "total": 0},
            "total": {"job_ids": [], "total": 1},
        },
        "status": "running",
        "user": "00000000-0000-0000-0000-000000000000",
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(
                cli,
                ["status", "-t", reana_token, "--json", "-v", "-w", response["name"]],
            )
            json_response = json.loads(result.output)
            assert result.exit_code == 0
            assert isinstance(json_response, list)
            assert len(json_response) == 1
            assert json_response[0]["name"] in response["name"]


@patch("reana_client.cli.workflow.workflow_create")
@patch("reana_client.cli.workflow.upload_files")
@patch("reana_client.cli.workflow.workflow_start")
def test_run(
    workflow_start_mock,
    upload_file_mock,
    workflow_create_mock,
    create_yaml_workflow_schema,
):
    """Test run command, if wrapped commands are called."""
    reana_workflow_schema = "reana.yaml"
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    reana_token = "000000"
    with runner.isolated_filesystem():
        with open(reana_workflow_schema, "w") as f:
            f.write(create_yaml_workflow_schema)
        runner.invoke(
            cli, ["run", "-t", reana_token, "-f", reana_workflow_schema],
        )
    assert workflow_create_mock.called is True
    assert upload_file_mock.called is True
    assert workflow_start_mock.called is True


def test_workflow_input_parameters():
    """Test if not existing input parameters from CLI are applied."""
    status_code = 200
    response = {
        "id": "d9304bdf-0d19-45d9-ae87-d5fd18059193",
        "name": "workflow.19",
        "type": "serial",
        "parameters": {
            "helloworld": "code/helloworld.py",
            "inputfile": "data/names.txt",
            "outputfile": "results/greetings.txt",
            "sleeptime": 2,
        },
    }
    env = {"REANA_SERVER_URL": "localhost"}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            parameter = "Debug"
            expected_message = "{0}, is not in reana.yaml".format(parameter)
            result = runner.invoke(
                cli,
                [
                    "start",
                    "-t",
                    reana_token,
                    "-w workflow.19",
                    "-p {0}=True".format(parameter),
                ],
            )
            assert expected_message in result.output


@pytest.mark.parametrize(
    "interactive_session_type",
    INTERACTIVE_SESSION_TYPES
    + [pytest.param("wrong-interactive-type", marks=pytest.mark.xfail)],
)
def test_open_interactive_session(interactive_session_type):
    """Test opening an interactive session."""
    status_code = 200
    workflow_id = "d9304bdf-0d19-45d9-ae87-d5fd18059193"
    response = {"path": "/{}".format(workflow_id)}
    reana_server_url = "http://localhost"
    env = {"REANA_SERVER_URL": reana_server_url}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = response
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            expected_message = "{reana_server_url}/{workflow_id}".format(
                reana_server_url=reana_server_url, workflow_id=workflow_id
            )
            result = runner.invoke(
                cli,
                [
                    "open",
                    "-t",
                    reana_token,
                    "-w",
                    workflow_id,
                    interactive_session_type,
                ],
            )
            assert expected_message in result.output


def test_close_interactive_session():
    """Test closing an interactive session."""
    status_code = 200
    workflow = "workflow.1"
    expected_message = (
        "Interactive session for workflow {} "
        "was successfully closed\n".format(workflow)
    )
    reana_server_url = "http://localhost"
    env = {"REANA_SERVER_URL": reana_server_url}
    mock_http_response, mock_response = Mock(), Mock()
    mock_http_response.status_code = status_code
    mock_response = expected_message
    reana_token = "000000"
    runner = CliRunner(env=env)
    with runner.isolation():
        with patch(
            "reana_client.api.client.current_rs_api_client",
            make_mock_api_client("reana-server")(mock_response, mock_http_response),
        ):
            result = runner.invoke(cli, ["close", "-t", reana_token, "-w", workflow])
            assert expected_message == result.output


def test_multiple_specifications(create_yaml_workflow_schema):
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    message = "ERROR: Found 2 REANA specification files " "(reana.yaml, reana.yml)."
    with runner.isolated_filesystem():
        with open("reana.yaml", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        with open("reana.yml", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        result = runner.invoke(cli, ["validate"])
        assert result.exit_code == 1
        assert message in result.output


def test_yml_ext_specification(create_yaml_workflow_schema):
    env = {"REANA_SERVER_URL": "localhost"}
    runner = CliRunner(env=env)
    message = "Valid REANA specification file"
    with runner.isolated_filesystem():
        with open("reana.yml", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        result = runner.invoke(cli, ["validate"])
        assert result.exit_code == 0
        assert message in result.output

    message = "ERROR: No REANA specification file (reana.yaml) found."
    with runner.isolated_filesystem():
        with open("reana.json", "w") as reana_schema:
            reana_schema.write(create_yaml_workflow_schema)
        result = runner.invoke(cli, ["validate"])
        assert result.exit_code != 0
        assert message in result.output



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2021 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client validate parameters tests."""

import pytest
import yaml
from click.testing import CliRunner

from reana_client.utils import cwl_load
from reana_client.validation.parameters import (
    SerialParameterValidator,
    validate_parameters,
)


def test_validate_parameters_cwl(
    create_cwl_yaml_workflow_schema,
    cwl_workflow_spec_step,
    cwl_workflow_spec_correct_input_param,
    cwl_workflow_spec_wrong_input_param,
    capsys,
):
    """Validate parameters for CWL workflows."""
    runner = CliRunner()
    workflow_type = "cwl"

    def get_loaded_yaml(step_spec, input_spec):
        with open("main.cwl", "w") as f:
            f.write(step_spec)
        with open("test.tool", "w") as g:
            g.write(input_spec)
        reana_yaml = yaml.load(create_cwl_yaml_workflow_schema, Loader=yaml.FullLoader)
        reana_yaml["workflow"]["specification"] = cwl_load(
            reana_yaml["workflow"].get("file")
        )
        return reana_yaml

    with runner.isolated_filesystem():
        reana_yaml = get_loaded_yaml(
            cwl_workflow_spec_step, cwl_workflow_spec_correct_input_param
        )
        validate_parameters(workflow_type, reana_yaml)
        captured = capsys.readouterr()
        assert "SUCCESS: Workflow operations appear valid." in captured.out

    # Wrong Input parameter used
    with runner.isolated_filesystem():
        reana_yaml = get_loaded_yaml(
            cwl_workflow_spec_step, cwl_workflow_spec_wrong_input_param
        )
        with pytest.raises(SystemExit) as exc_info:
            validate_parameters(workflow_type, reana_yaml)
        assert (
            "Step is missing required parameter 'xoutputfile'" in exc_info.value.args[0]
        )

    # Wrong file path used
    reana_yaml = yaml.load(create_cwl_yaml_workflow_schema, Loader=yaml.FullLoader)
    with pytest.raises(SystemExit) as exc_info:
        validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert "ERROR: Workflow path main.cwl is not valid." in captured.err


def test_validate_parameters_serial(create_yaml_workflow_schema, capsys):
    """Validate parameters for Serial workflows."""
    reana_yaml = yaml.load(create_yaml_workflow_schema, Loader=yaml.FullLoader)
    workflow_type = "serial"
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert "Workflow parameters and commands appear valid." in captured.out

    # Input parameter not being used
    reana_yaml["inputs"]["parameters"]["foo"] = "foo"
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: REANA input parameter "foo" does not seem to be used' in captured.out
    )
    del reana_yaml["inputs"]["parameters"]["foo"]

    # Escaped env vars don't trigger validation warning
    reana_yaml["workflow"]["specification"]["steps"].append(
        {"commands": [r"$\{SHELL\} -c echo foo"]}
    )
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert "Workflow parameters and commands appear valid." in captured.out

    # Parameter in unnamed step not defined
    reana_yaml["workflow"]["specification"]["steps"].append(
        {"commands": ["python ${foo} --bar"]}
    )
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: Serial parameter "foo" found on step "2" is not defined in input parameters'
        in captured.out
    )

    # Parameter not defined in named step
    reana_yaml["workflow"]["specification"]["steps"].append(
        {"commands": ["python ${bar} --foo"], "name": "baz"}
    )
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: Serial parameter "bar" found on step "baz" is not defined in input parameters'
        in captured.out
    )

    # Parameter not defined in two named steps
    reana_yaml["workflow"]["specification"]["steps"].append(
        {"commands": ["python ${bar} --foo"], "name": "qux"}
    )
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert 'WARNING: Serial parameter "bar" found on steps' in captured.out
    assert "baz, qux" in captured.out or "qux, baz" in captured.out
    assert "is not defined in input parameters" in captured.out


def test_validate_parameters_yadage(yadage_workflow_spec_loaded, capsys):
    """Validate parameters for Yadage workflows."""
    reana_yaml = yadage_workflow_spec_loaded
    workflow_type = "yadage"
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert "Workflow parameters and commands appear valid." in captured.out

    # REANA input parameter not being used.
    reana_yaml["inputs"]["parameters"]["qux"] = "qux_val"
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: REANA input parameter "qux" does not seem to be used' in captured.out
    )
    del reana_yaml["inputs"]["parameters"]["qux"]

    # Yadage parameter not being used.
    reana_yaml["workflow"]["specification"]["stages"][0]["scheduler"][
        "parameters"
    ].append(
        {
            "key": "qux",
            "value": {
                "step": "init",
                "output": "qux",
                "expression_type": "stage-output-selector",
            },
        }
    )
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: Yadage input parameter "qux" found on step "gendata" does not seem to be used.'
        in captured.out
    )
    reana_yaml["workflow"]["specification"]["stages"][0]["scheduler"][
        "parameters"
    ].pop()

    # Parameter not defined in step
    process = reana_yaml["workflow"]["specification"]["stages"][0]["scheduler"]["step"][
        "process"
    ]
    process["script"] += " && ./run-job {my_job}"

    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: Yadage parameter "my_job" found on step "gendata" is not defined in input parameters'
        in captured.out
    )

    # Parameter not defined in two steps
    process = reana_yaml["workflow"]["specification"]["stages"][1]["scheduler"]["step"][
        "process"
    ]
    process["script"] += " && ./run-job {my_job}"

    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert 'WARNING: Yadage parameter "my_job" found on steps' in captured.out
    assert "gendata, fitdata" in captured.out or "fitdata, gendata" in captured.out
    assert "is not defined in input parameters" in captured.out

    # Yadage parameter defined in sub-step not being used.
    subworkflow = reana_yaml["workflow"]["specification"]["stages"][2]["scheduler"]
    subworkflow["workflow"]["stages"][0]["scheduler"]["parameters"].append(
        {"key": "subfoo", "value": "subfoo_val"}
    )
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: Yadage input parameter "subfoo" found on step "nested_step" does not seem to be used.'
        in captured.out
    )

    # Use previous parameter in command
    process = subworkflow["workflow"]["stages"][0]["scheduler"]["step"]["process"]
    process["script"] += " && go run {subfoo}"
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert "subfoo" not in captured.out

    # Parameter not defined in sub-step
    process = subworkflow["workflow"]["stages"][0]["scheduler"]["step"]["process"]
    process["script"] += " && go run {subbar}"
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: Yadage parameter "subbar" found on step "nested_step" is not defined in input parameters'
        in captured.out
    )

    # Parameter defined in one stage, but forgotten in the other one
    process = reana_yaml["workflow"]["specification"]["stages"][1]["scheduler"]["step"][
        "process"
    ]
    process["script"] += " && ./run-job {foo}"
    validate_parameters(workflow_type, reana_yaml)
    captured = capsys.readouterr()
    assert (
        'WARNING: Yadage parameter "foo" found on step "fitdata" is not defined in input parameters.'
        in captured.out
    )


@pytest.mark.parametrize(
    "commands, step, warning",
    [
        (["python foo.py"], "gendata", ""),
        (
            ["python foo.py", "sudo python foo.py"],
            "fitdata",
            '"sudo" found in step "fitdata" might be dangerous.',
        ),
        (
            ['echo "hello world!" && sudo python foo.py'],
            "fitdata",
            '"sudo" found in step "fitdata" might be dangerous.',
        ),
        (
            ["cd /foo && npm install", 'echo "hello world!"'],
            "installation",
            '"cd /" found in step "installation" might be dangerous.',
        ),
        (["sudo npm install"], None, '"sudo" might be dangerous.',),
    ],
)
def test_validate_dangerous_operations(commands, step, warning, capsys):
    """Validate if dangerous operations in a command trigger a warning."""
    validator = SerialParameterValidator({})
    validator._validate_dangerous_operations(commands, step=step)
    warnings = validator.operations_warnings
    assert warning in (warnings.pop()["message"] if warnings else "")



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Common click options."""

import functools
import json
import os
import shlex
import sys

import click

from reana_client.utils import workflow_uuid_or_name
from reana_client.config import ERROR_MESSAGES, RUN_STATUSES
from reana_commons.errors import MissingAPIClientConfiguration


def add_access_token_options(func):
    """Add access token related options to click commands."""

    @click.option(
        "-t",
        "--access-token",
        default=os.getenv("REANA_ACCESS_TOKEN", None),
        callback=access_token_check,
        help="Access token of the current user.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def human_readable_or_raw_option(func):
    """Add human readable option to click commands."""

    @click.option(
        "-h",
        "--human-readable",
        "human_readable_or_raw",
        is_flag=True,
        default=False,
        callback=lambda ctx, param, value: "human_readable" if value else "raw",
        help="Show disk size in human readable format.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def access_token_check(ctx, _, access_token):
    """Check if access token is present."""
    if not access_token:
        click.echo(
            click.style(ERROR_MESSAGES["missing_access_token"], fg="red"), err=True
        )
        ctx.exit(1)
    else:
        return access_token


def check_connection(func):
    """Check if connected to any REANA cluster."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from reana_client.utils import get_api_url

        api_url = get_api_url()
        if not api_url:
            click.secho(
                "REANA client is not connected to any REANA cluster.",
                fg="red",
                err=True,
            )
            sys.exit(1)
        return func(*args, **kwargs)

    return wrapper


def add_workflow_option(func):
    """Add workflow related option to click commands."""

    @click.option(
        "-w",
        "--workflow",
        default=os.environ.get("REANA_WORKON", None),
        callback=workflow_uuid_or_name,
        help="Name or UUID of the workflow. Overrides value of "
        "REANA_WORKON environment variable.",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def add_pagination_options(func):
    """Add pagination related options to click commands."""

    @click.option(
        "--page",
        default=1,
        type=int,
        help="Results page number (to be used with --size).",
    )
    @click.option(
        "--size", type=int, help="Size of results per page (to be used with --page).",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def parse_format_parameters(_format):
    """Return parsed filter parameters."""
    try:
        parsed_filters = []
        filters = " ".join(_format).replace(",", " ")
        for item in shlex.split(filters):
            if "=" in item:
                filter_item = {
                    "column_name": item.split("=")[0].lower(),
                    "column_value": item.split("=")[1],
                }
            else:
                filter_item = {"column_name": item.lower(), "column_value": None}
            parsed_filters.append(filter_item)
        return parsed_filters
    except ValueError as e:
        click.echo(
            click.style("Wrong filter format \n{0}".format(e.message), fg="red"),
            err=True,
        )


def parse_filter_parameters(filters, filter_names):
    """Return parsed filter parameters."""
    try:
        status_filters = []
        search_filters = {}
        filters = list(filters)
        for item in filters:
            if "=" in item:
                filter_name = item.split("=")[0].lower()
                filter_value = item.split("=")[1]
                if filter_name == "status" and filter_name in filter_names:
                    if filter_value in RUN_STATUSES:
                        status_filters.append(filter_value)
                    else:
                        click.secho(
                            "==> ERROR: Input status value {} is not valid. ".format(
                                filter_value
                            ),
                            err=True,
                            fg="red",
                        ),
                        sys.exit(1)
                elif filter_name in filter_names:
                    if search_filters.get(filter_name):
                        search_filters[filter_name].append(filter_value)
                    else:
                        search_filters[filter_name] = [filter_value]
                else:
                    click.secho(
                        "==> ERROR: Filter {} is not valid.".format(filter_name),
                        err=True,
                        fg="red",
                    ),
                    sys.exit(1)
            else:
                raise click.BadParameter(
                    "Wrong input format. Please use --filter filter_name=filter_value"
                )
        if not search_filters:
            search_filters = None
        else:
            search_filters = json.dumps(search_filters)
        return status_filters, search_filters
    except ValueError as e:
        click.echo(
            click.style("Wrong filter format \n{0}".format(e.message), fg="red"),
            err=True,
        )


def format_data(parsed_filters, headers, tablib_data):
    """Return filtered data."""
    parsed_filters = [i for i in parsed_filters if i["column_name"] in headers]
    column_headers = [i["column_name"] for i in parsed_filters] or None
    tablib_data = tablib_data.subset(rows=None, cols=column_headers)
    tablib_data = json.loads(tablib_data.export("json"))
    filtered_data = list(tablib_data)
    for item in filtered_data:
        for filter_ in parsed_filters:
            if (
                filter_["column_value"] is not None
                and filter_["column_value"] != item[filter_["column_name"]]
            ):
                tablib_data.remove(item)
                break
    return tablib_data, column_headers or []


def format_session_uri(reana_server_url, path, access_token):
    """Format interactive session URI."""
    return "{reana_server_url}{path}?token={access_token}".format(
        reana_server_url=reana_server_url, path=path, access_token=access_token
    )


def get_formatted_progress(progress):
    """Return workflow progress in format of finished/total jobs."""
    total_jobs = progress.get("total", {}).get("total") or "-"
    finished_jobs = progress.get("finished", {}).get("total") or "-"
    return "{0}/{1}".format(finished_jobs, total_jobs)


def validate_workflow_name(ctx, _, workflow_name):
    """Validate workflow name."""
    not_allowed_characters = ["."]
    if workflow_name:
        for item in not_allowed_characters:
            if item in workflow_name:
                click.echo(
                    click.style(
                        "Workflow name {} contains illegal "
                        'character "{}"'.format(workflow_name, item),
                        fg="red",
                    ),
                    err=True,
                )
                sys.exit(1)
    return workflow_name


def key_value_to_dict(ctx, param, value):
    """Convert tuple params to dictionary. e.g `(foo=bar)` to `{'foo': 'bar'}`.

    :param options: A tuple with CLI operational options.
    :returns: A dictionary representation of the given options.
    """
    try:
        return dict(op.split("=") for op in value)
    except ValueError:
        click.secho(
            '==> ERROR: Input parameter "{0}" is not valid. '
            'It must follow format "param=value".'.format(" ".join(value)),
            err=True,
            fg="red",
        ),
        sys.exit(1)


def requires_environments(ctx, param, value):
    """Require passing ``--environments`` flag."""
    if value and not ctx.params.get("environments"):
        click.secho(
            "==> ERROR: `{}` flag requires `--environments` flag.".format(
                param.opts[0]
            ),
            err=True,
            fg="red",
        )
        sys.exit(1)
    return value


class NotRequiredIf(click.Option):
    """Allow only one of two arguments to be missing."""

    def __init__(self, *args, **kwargs):
        """."""
        self.not_required_if = kwargs.pop("not_required_if")
        assert self.not_required_if, "'not_required_if' parameter required"
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        """Overwritten click method."""
        argument_present = self.name in opts
        other_argument_present = self.not_required_if in opts
        if not argument_present and not other_argument_present:
            click.echo(
                click.style(
                    "At least one of the options: `{}` or `{}` is required\n".format(
                        self.name, self.not_required_if
                    ),
                    fg="red",
                )
                + ctx.get_help(),
                err=True,
            )
            sys.exit(1)

        return super(NotRequiredIf, self).handle_parse_result(ctx, opts, args)


def output_user_friendly_logs(workflow_logs, steps):
    """Output workflow logs in a user-friendly manner.

    :param workflow_logs: Dictionary representing the workflow logs as they
        are returned from the REST API.
    :param steps: List of steps to show logs for.
    """
    # Colors matching the status colors of REANA-UI
    status_to_color_mapping = {
        "running": "bright_blue",
        "finished": "green",
        "failed": "red",
    }
    key_to_description_mapping = {
        "workflow_uuid": "Workflow ID",
        "compute_backend": "Compute backend",
        "backend_job_id": "Job ID",
        "docker_img": "Docker image",
        "cmd": "Command",
        "status": "Status",
        "started_at": "Started",
        "finished_at": "Finished",
    }
    leading_mark = "==>"

    # REANA Workflow Engine logs
    if workflow_logs.get("workflow_logs"):
        click.secho(
            "{} Workflow engine logs".format(leading_mark), bold=True, fg="yellow"
        )
        click.echo(workflow_logs["workflow_logs"])

    if workflow_logs.get("engine_specific"):
        click.echo("\n")
        click.secho("{} Engine internal logs".format(leading_mark), fg="yellow")
        click.secho(workflow_logs["engine_specific"])

    returned_step_names = set(
        workflow_logs["job_logs"][item]["job_name"]
        for item in workflow_logs["job_logs"].keys()
    )
    if steps:
        missing_steps = set(steps).difference(returned_step_names)
        if missing_steps:
            click.echo(
                click.style(
                    "The logs of step(s) {} were not found, "
                    "check for spelling mistakes in the step "
                    "names.".format(",".join(missing_steps)),
                    fg="red",
                )
            )
    # Job logs
    if workflow_logs["job_logs"]:
        click.echo("\n")
        click.secho("{} Job logs".format(leading_mark), bold=True, fg="yellow")
    for job_id, logs_info in workflow_logs["job_logs"].items():
        if logs_info:
            job_name_or_id = logs_info["job_name"] or job_id
            click.secho(
                "{0} Step: {1}".format(leading_mark, job_name_or_id),
                bold=True,
                fg=status_to_color_mapping.get(logs_info["status"]),
            )
            logs_output = logs_info["logs"]
            # extract already used fields
            del logs_info["logs"]
            del logs_info["job_name"]

            for key, value in logs_info.items():
                if value:
                    title = click.style(
                        "{mark} {description}:".format(
                            mark=leading_mark,
                            description=key_to_description_mapping[key],
                        ),
                        fg=status_to_color_mapping.get(logs_info["status"]),
                    )
                    click.secho("{title} {value}".format(title=title, value=value))
            # show actual log content
            if logs_output:
                click.secho(
                    "{mark} Logs:".format(mark=leading_mark),
                    fg=status_to_color_mapping.get(logs_info["status"]),
                )
                click.secho(logs_output)
            else:
                click.secho(
                    "{0} Step {1} emitted no logs.".format(
                        leading_mark, job_name_or_id
                    ),
                    bold=True,
                )



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018, 2019 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client"""

from __future__ import absolute_import, print_function

import os
import re

from setuptools import find_packages, setup

readme = open("README.rst").read()
history = open("CHANGES.rst").read()

tests_require = [
    "pytest-reana>=0.8.0a5,<0.9.0",
    'google-auth==1.24.0 ; python_version=="2.7"',
    'oauthlib==3.1.0; python_version=="2.7"',
    'pluggy==0.13.1; python_version=="2.7"',
    'websocket-client==0.59.0; python_version=="2.7"',
]

extras_require = {
    "docs": ["Sphinx>=1.5.1", "sphinx-rtd-theme>=0.1.9", "sphinx-click>=1.0.4",],
    "tests": tests_require,
}

extras_require["all"] = []
for key, reqs in extras_require.items():
    if ":" == key[0]:
        continue
    extras_require["all"].extend(reqs)

setup_requires = [
    "pytest-runner>=2.7",
]

install_requires = [
    "PyYAML>=5.1",
    'typing>=3.7.4 ; python_version=="2.7"',  # workaround for CWL deps
    'click>=7 ; python_version>="3"',
    'click==7.0 ; python_version=="2.7"',
    "cwltool==1.0.20191022103248",
    'cwl-utils==0.5 ; python_version>="3"',
    "jsonpointer>=2.0",
    "reana-commons[yadage]>=0.8.0a17,<0.9.0",
    "six>=1.12.0",
    "tablib>=0.12.1,<0.13",
    "werkzeug>=0.14.1",
    'websocket-client==0.59.0 ; python_version=="2.7"',  # pin due to py2 support drop
]

packages = find_packages()


# Get the version string. Cannot be done with import!
with open(os.path.join("reana_client", "version.py"), "rt") as f:
    version = re.search(r'__version__\s*=\s*"(?P<version>.*)"\n', f.read()).group(
        "version"
    )

setup(
    name="reana-client",
    version=version,
    description=__doc__,
    long_description=readme + "\n\n" + history,
    author="REANA",
    author_email="info@reana.io",
    url="https://github.com/reanahub/reana-client",
    packages=["reana_client",],
    zip_safe=False,
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "reana-client = reana_client.cli:cli",
            "reana-cwl-runner = reana_client.cli.cwl_runner:cwl_runner",
        ],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2017, 2018 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""REANA client debugging commands."""

import logging
import traceback

import click
from reana_client.cli.utils import add_access_token_options, check_connection
from reana_client.version import __version__


@click.group(help="Configuration commands")
def configuration_group():
    """Configuration commands."""
    pass


@configuration_group.command("ping")
@click.pass_context
@add_access_token_options
@check_connection
def ping(ctx, access_token):  # noqa: D301
    """Check connection to REANA server.

    The `ping` command allows to test connection to REANA server.

    Examples: \n
    \t $ reana-client ping
    """
    try:
        from reana_client.api.client import ping as rs_ping
        from reana_client.utils import get_api_url

        logging.info("Connecting to {0}".format(get_api_url()))
        response = rs_ping(access_token)
        msg_color = "red" if response.get("error") else "green"
        click.echo(
            click.style(
                "REANA server: {0}\n"
                "REANA server version: {1}\n"
                "REANA client version: {2}\n"
                "Authenticated as: {3} <{4}>\n"
                "Status: {5}".format(
                    get_api_url(),
                    response.get("reana_server_version", ""),
                    __version__,
                    response.get("full_name", ""),
                    response.get("email"),
                    response.get("status"),
                ),
                fg=msg_color,
            )
        )
        logging.debug("Server response:\n{}".format(response))

    except Exception as e:
        logging.debug(traceback.format_exc())
        logging.debug(str(e))
        error_msg = (
            "Could not connect to the selected REANA cluster "
            "server at {0}:\n{1}".format(get_api_url(), e)
        )
        click.echo(click.style(error_msg, fg="red"), err=True)
        ctx.exit(1)


@configuration_group.command("version")
@click.pass_context
def version(ctx):  # noqa: D301
    """Show version.

    The `version` command shows REANA client version.

    Examples: \n
    \t $ reana-client version
    """
    click.echo(__version__)



# -*- coding: utf-8 -*-
#
# This file is part of REANA.
# Copyright (C) 2021 CERN.
#
# REANA is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""REANA client environment validation."""

import sys
import logging

import requests

from reana_commons.config import WORKFLOW_RUNTIME_USER_GID, WORKFLOW_RUNTIME_USER_UID

from reana_client.errors import EnvironmentValidationError
from reana_client.config import (
    DOCKER_REGISTRY_INDEX_URL,
    ENVIRONMENT_IMAGE_SUSPECTED_TAGS_VALIDATOR,
    GITLAB_CERN_REGISTRY_INDEX_URL,
    GITLAB_CERN_REGISTRY_PREFIX,
)
from reana_client.printer import display_message


def validate_environment(reana_yaml, pull=False):
    """Validate environments in REANA specification file according to workflow type.

    :param reana_yaml: Dictionary which represents REANA specifications file.
    :param pull: If true, attempt to pull remote environment image to perform GID/UID validation.
    """

    def build_validator(workflow):
        workflow_type = workflow["type"]
        if workflow_type == "serial":
            workflow_steps = workflow["specification"]["steps"]
            return SerialEnvironmentValidator(workflow_steps=workflow_steps, pull=pull)
        if workflow_type == "yadage":
            workflow_steps = workflow["specification"]["stages"]
            return YadageEnvironmentValidator(workflow_steps=workflow_steps, pull=pull)
        if workflow_type == "cwl":
            workflow_file = workflow.get("file")
            return CWLEnvironmentValidator(workflow_file=workflow_file, pull=pull)

    workflow = reana_yaml["workflow"]
    validator = build_validator(workflow)
    validator.validate()
    validator.display_messages()


class EnvironmentValidatorBase:
    """REANA workflow environments validation base class."""

    def __init__(self, workflow_steps=None, workflow_file=None, pull=False):
        """Validate environments in REANA workflow.

        :param workflow_steps: List of dictionaries which represents different steps involved in workflow.
        :param workflow_file: Path to workflow specification.
        :param pull: If true, attempt to pull remote environment image to perform GID/UID validation.
        """
        self.workflow_steps = workflow_steps
        self.workflow_file = workflow_file
        self.pull = pull
        self.validated_images = set()
        self.messages = []

    def validate(self):
        """Validate REANA workflow environments."""
        try:
            self.validate_environment()
        except EnvironmentValidationError as e:
            self.messages.append({"type": "error", "message": str(e)})
            self.display_messages()
            sys.exit(1)

    def validate_environment(self):
        """Validate environments in REANA workflow."""
        raise NotImplementedError

    def display_messages(self):
        """Display messages in console."""
        for msg in self.messages:
            display_message(
                msg["message"], msg_type=msg["type"], indented=True,
            )

    def _validate_environment_image(self, image, kubernetes_uid=None):
        """Validate image environment.

        :param image: Full image name with tag if specified.
        :param kubernetes_uid: Kubernetes UID defined in workflow spec.
        """

        if image not in self.validated_images:
            image_name, image_tag = self._validate_image_tag(image)
            exists_locally, _ = self._image_exists(image_name, image_tag)
            if exists_locally or self.pull:
                uid, gids = self._get_image_uid_gids(image_name, image_tag)
                self._validate_uid_gids(uid, gids, kubernetes_uid=kubernetes_uid)
            else:
                self.messages.append(
                    {
                        "type": "warning",
                        "message": "UID/GIDs validation skipped, specify `--pull` to enable it.",
                    }
                )
            self.validated_images.add(image)

    def _image_exists(self, image, tag):
        """Verify if image exists locally or remotely.

        :returns: A tuple with two boolean values: image exists locally, image exists remotely.
        """

        image_exists_remotely = (
            self._image_exists_in_gitlab_cern
            if image.startswith(GITLAB_CERN_REGISTRY_PREFIX)
            else self._image_exists_in_dockerhub
        )

        exists_locally = self._image_exists_locally(image, tag)
        exists_remotely = image_exists_remotely(image, tag)

        if not any([exists_locally, exists_remotely]):
            raise EnvironmentValidationError(
                "Environment image {} does not exist locally or remotely.".format(
                    self._get_full_image_name(image, tag)
                )
            )
        return exists_locally, exists_remotely

    def _validate_uid_gids(self, uid, gids, kubernetes_uid=None):
        """Check whether container UID and GIDs are valid."""
        if WORKFLOW_RUNTIME_USER_GID not in gids:
            if kubernetes_uid is None:
                raise EnvironmentValidationError(
                    "Environment image GID must be {}. GIDs {} were found.".format(
                        WORKFLOW_RUNTIME_USER_GID, gids
                    )
                )
            else:
                self.messages.append(
                    {
                        "type": "warning",
                        "message": "Environment image GID is recommended to be {}. GIDs {} were found.".format(
                            WORKFLOW_RUNTIME_USER_GID, gids
                        ),
                    }
                )
        if kubernetes_uid is not None:
            if kubernetes_uid != uid:
                self.messages.append(
                    {
                        "type": "warning",
                        "message": "`kubernetes_uid` set to {}. UID {} was found.".format(
                            kubernetes_uid, uid
                        ),
                    }
                )
        elif uid != WORKFLOW_RUNTIME_USER_UID:
            self.messages.append(
                {
                    "type": "info",
                    "message": "Environment image uses UID {} but will run as UID {}.".format(
                        uid, WORKFLOW_RUNTIME_USER_UID
                    ),
                }
            )

    def _validate_image_tag(self, image):
        """Validate if image tag is valid."""
        image_name, image_tag = "", ""
        message = {
            "type": "success",
            "message": "Environment image {} has the correct format.".format(image),
        }
        if ":" in image:
            environment = image.split(":", 1)
            image_name, image_tag = environment[0], environment[-1]
            if ":" in image_tag:
                raise EnvironmentValidationError(
                    "Environment image {} has invalid tag '{}'".format(
                        image_name, image_tag
                    )
                )
            elif image_tag in ENVIRONMENT_IMAGE_SUSPECTED_TAGS_VALIDATOR:
                message = {
                    "type": "warning",
                    "message": "Using '{}' tag is not recommended in {} environment image.".format(
                        image_tag, image_name
                    ),
                }
        else:
            message = {
                "type": "warning",
                "message": "Environment image {} does not have an explicit tag.".format(
                    image
                ),
            }
            image_name = image

        self.messages.append(message)
        return image_name, image_tag

    def _image_exists_locally(self, image, tag):
        """Verify if image exists locally."""
        full_image = self._get_full_image_name(image, tag or "latest")
        local_images = self._get_local_docker_images()
        if full_image in local_images:
            self.messages.append(
                {
                    "type": "success",
                    "message": "Environment image {} exists locally.".format(
                        full_image
                    ),
                }
            )
            return True
        else:
            self.messages.append(
                {
                    "type": "warning",
                    "message": "Environment image {} does not exist locally.".format(
                        full_image
                    ),
                }
            )
            return False

    def _image_exists_in_gitlab_cern(self, image, tag):
        """Verify if image exists in GitLab CERN."""
        # Remove registry prefix
        prefixed_image = image
        full_prefixed_image = self._get_full_image_name(image, tag or "latest")
        image = image.split("/", 1)[-1]
        # Encode image name slashes
        remote_registry_url = GITLAB_CERN_REGISTRY_INDEX_URL.format(
            image=requests.utils.quote(image, safe="")
        )
        try:
            # FIXME: if image is private we can't access it, we'd
            # need to pass a GitLab API token generated from the UI.
            response = requests.get(remote_registry_url)
        except requests.exceptions.RequestException as e:
            logging.error(e)
            self.messages.append(
                {
                    "type": "error",
                    "message": "Something went wrong when querying {}".format(
                        remote_registry_url
                    ),
                }
            )
            return False

        if not response.ok:
            msg = response.json().get("message")
            self.messages.append(
                {
                    "type": "warning",
                    "message": "Existence of environment image {} in GitLab CERN could not be verified: {}".format(
                        self._get_full_image_name(prefixed_image, tag), msg
                    ),
                }
            )
            return False
        else:
            # If not tag was set, use `latest` (default) to verify.
            tag = tag or "latest"
            tag_exists = any(
                tag_dict["name"] == tag for tag_dict in response.json()[0].get("tags")
            )
            if tag_exists:
                self.messages.append(
                    {
                        "type": "success",
                        "message": "Environment image {} exists in GitLab CERN.".format(
                            full_prefixed_image
                        ),
                    }
                )
                return True
            else:
                self.messages.append(
                    {
                        "type": "warning",
                        "message": 'Environment image {} in GitLab CERN does not exist: Tag "{}" missing.'.format(
                            full_prefixed_image, tag
                        ),
                    }
                )
                return False

    def _image_exists_in_dockerhub(self, image, tag):
        """Verify if image exists in DockerHub."""
        full_image = self._get_full_image_name(image, tag or "latest")
        docker_registry_url = DOCKER_REGISTRY_INDEX_URL.format(image=image, tag=tag)
        # Remove traling slash if no tag was specified
        if not tag:
            docker_registry_url = docker_registry_url[:-1]
        try:
            response = requests.get(docker_registry_url)
        except requests.exceptions.RequestException as e:
            logging.error(e)
            self.messages.append(
                {
                    "type": "error",
                    "message": "Something went wrong when querying {}".format(
                        docker_registry_url
                    ),
                }
            )
            return False

        if not response.ok:
            if response.status_code == 404:
                msg = response.text
                self.messages.append(
                    {
                        "type": "warning",
                        "message": "Environment image {} does not exist in Docker Hub: {}".format(
                            full_image, msg
                        ),
                    }
                )
                return False
            else:
                self.messages.append(
                    {
                        "type": "warning",
                        "message": "==> WARNING: Existence of environment image {} in Docker Hub could not be verified. Status code: {} {}".format(
                            full_image, response.status_code, response.reason,
                        ),
                    }
                )
                return False
        else:
            self.messages.append(
                {
                    "type": "success",
                    "message": "Environment image {} exists in Docker Hub.".format(
                        full_image
                    ),
                }
            )
            return True

    def _get_image_uid_gids(self, image, tag):
        """Obtain environment image UID and GIDs.

        :returns: A tuple with UID and GIDs.
        """
        from reana_client.utils import run_command

        # Check if docker is installed.
        run_command("docker version", display=False, return_output=True)
        # Run ``id``` command inside the container.
        uid_gid_output = run_command(
            'docker run -i -t --rm {} sh -c "/usr/bin/id -u && /usr/bin/id -G"'.format(
                self._get_full_image_name(image, tag)
            ),
            display=False,
            return_output=True,
        )
        ids = uid_gid_output.splitlines()
        uid, gids = (
            int(ids[-2]),
            [int(gid) for gid in ids[-1].split()],
        )
        return uid, gids

    def _get_full_image_name(self, image, tag=None):
        """Return full image name with tag if is passed."""
        return "{}{}".format(image, ":{}".format(tag) if tag else "")

    def _get_local_docker_images(self):
        """Return a list with local docker images."""
        from reana_client.utils import run_command

        # Check if docker is installed.
        run_command("docker version", display=False, return_output=True)
        docker_images = run_command(
            'docker images --format "{{ .Repository }}:{{ .Tag }}"',
            display=False,
            return_output=True,
        )
        return docker_images.splitlines()


class SerialEnvironmentValidator(EnvironmentValidatorBase):
    """REANA serial workflow environments validation."""

    def validate_environment(self):
        """Validate environments in REANA serial workflow."""
        for step in self.workflow_steps:
            image = step["environment"]
            kubernetes_uid = step.get("kubernetes_uid")
            self._validate_environment_image(image, kubernetes_uid=kubernetes_uid)


class YadageEnvironmentValidator(EnvironmentValidatorBase):
    """REANA yadage workflow environments validation."""

    def _extract_steps_environments(self):
        """Extract environments yadage workflow steps."""

        def traverse_yadage_workflow(stages):
            environments = []
            for stage in stages:
                if "workflow" in stage["scheduler"]:
                    nested_stages = stage["scheduler"]["workflow"].get("stages", {})
                    environments += traverse_yadage_workflow(nested_stages)
                else:
                    environments.append(stage["scheduler"]["step"]["environment"])
            return environments

        return traverse_yadage_workflow(self.workflow_steps)

    def validate_environment(self):
        """Validate environments in REANA yadage workflow."""

        def _check_environment(environment):
            image = "{}{}".format(
                environment["image"],
                ":{}".format(environment["imagetag"])
                if "imagetag" in environment
                else "",
            )
            k8s_uid = next(
                (
                    resource["kubernetes_uid"]
                    for resource in environment.get("resources", [])
                    if "kubernetes_uid" in resource
                ),
                None,
            )
            self._validate_environment_image(image, kubernetes_uid=k8s_uid)

        steps_environments = self._extract_steps_environments()
        for environment in steps_environments:
            if environment["environment_type"] != "docker-encapsulated":
                raise EnvironmentValidationError(
                    'The only Yadage environment type supported is "docker-encapsulated". Found "{}".'.format(
                        environment["environment_type"]
                    )
                )
            else:
                _check_environment(environment)


class CWLEnvironmentValidator(EnvironmentValidatorBase):
    """REANA CWL workflow environments validation."""

    def validate_environment(self):
        """Validate environments in REANA CWL workflow."""

        try:
            import cwl_utils.parser_v1_0 as cwl_parser
            from cwl_utils.docker_extract import traverse
        except ImportError as e:
            display_message(
                "Cannot validate environment. Please install reana-client on Python 3+ to enable environment validation for CWL workflows.",
                msg_type="error",
                indented=True,
            )
            raise e

        top = cwl_parser.load_document(self.workflow_file)

        for image in traverse(top):
            self._validate_environment_image(image)
